r"""
.. note:: Please read the Validation section below.

Definition
----------
This model calculates the structure factor of a polyelectrolyte solution with
the RPA expression derived by Borue and Erukhimovich\ [#Borue]_.  Note however
that the fitting procedure here does not follow the notation in that reference
as 's' and 't' are **not** decoupled. Instead the scattering intensity $I(q)$
is calculated as

.. math::

    I(q) &= K\frac{q^2+k^2}{4\pi L_b\alpha ^2}
        \frac{1}{1+r_{0}^4(q^2+k^2)(q^2-12hC_a/b^2)} + \text{background} \\
    k^2 &= 4\pi L_b(2C_s + \alpha C_a) \\
    r_{0}^2 &= \frac{b}{\alpha \sqrt{C_a 48\pi L_b}}

where

$K$ is the contrast factor for the polymer which is defined differently than in
other models and is given in barns where 1 $barn = 10^{-24}$ $cm^2$.  $K$ is
defined as:

.. math::

    K &= a^2 \\
    a &= b_p - (v_p/v_s) b_s

where:

- $b_p$ and $b_s$ are **sum of the scattering lengths of the atoms**
  constituting the polymer monomer and the solvent molecules, respectively.

- $v_p$ and $v_s$ are the partial molar volume of the polymer and the
  solvent, respectively.

- $L_b$ is the Bjerrum length (|Ang|) - **Note:** This parameter needs to be
  kept constant for a given solvent and temperature!

- $h$ is the virial parameter (|Ang^3|) - **Note:** See [#Borue]_ for the
  correct interpretation of this parameter.  It incorporates second and third
  virial coefficients and can be *negative*.

- $b$ is the monomer length (|Ang|).

- $C_s$ is the concentration of monovalent salt(1/|Ang^3| - internally
  converted from mol/L).

- $\alpha$ is the degree of ionization (the ratio of charged monomers to the
  total number of monomers)

- $C_a$ is the polymer molar concentration (1/|Ang^3| - internally converted
  from mol/L)

- $\text{background}$ is the incoherent background.

For 2D data the scattering intensity is calculated in the same way as 1D,
where the $\vec q$ vector is defined as

.. math::

    q = \sqrt{q_x^2 + q_y^2}

Validation
----------

As of the last revision, this code is believed to be correct.  However it
needs further validation and should be used with caution at this time.  The
history of this code goes back to a 1998 implementation. It was recently noted
that in that implementation, while both the polymer concentration and salt
concentration were converted from experimental units of mol/L to more
dimensionally useful units of 1/|Ang^3|, only the converted version of the
polymer concentration was actually being used in the calculation while the
unconverted salt concentration (still in apparent units of mol/L) was being
used.  This was carried through to Sasmodels as used for SasView 4.1 (though
the line of code converting the salt concentration to the new units was removed
somewhere along the line). Simple dimensional analysis of the calculation shows
that the converted salt concentration should be used, which the original code
suggests was the intention, so this has now been corrected (for SasView 4.2).
Once better validation has been performed this note will be removed.

References
----------

For further details, see [#Joanny1990]_, [#Moussaid1993]_, [#Raphael1990]_.

.. [#Borue] V Y Borue, I Y Erukhimovich, *Macromolecules*, 21 (1988) 3240
.. [#Joanny1990] J F Joanny, L Leibler, *Journal de Physique*, 51 (1990) 545
.. [#Moussaid1993] A Moussaid, F Schosseler, J P Munch, S Candau,
   *J. Journal de Physique II France*, 3 (1993) 573
.. [#Raphael1990] E Raphael, J F Joanny, *Europhysics Letters*, 11 (1990) 179

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Paul Butler **Date:** September 25, 2018
* **Last Reviewed by:** Paul Butler **Date:** September 25, 2018
"""

import numpy as np
from numpy import inf, pi, sqrt

name = "be_polyelectrolyte"
title = "Polyelectrolyte with the RPA expression derived by Borue and Erukhimovich"
description = """
            Evaluate
            F(x) = K 1/(4 pi Lb (alpha)^(2)) (q^(2)+k2)/(1+(r02)^(2))
                 (q^(2)+k2) (q^(2)-(12 h C/b^(2)))

            has 3 internal parameters :
                   The inverse Debye Length: K2 = 4 pi Lb (2 Cs+alpha C)
                   r02 =1/alpha/Ca^(0.5) (B/(48 pi Lb)^(0.5))
                   Ca = 6.022136e-4 C
            """
category = "shape-independent"

# pylint: disable=bad-whitespace, line-too-long
#   ["name", "units", default, [lower, upper], "type", "description"],
parameters = [
    ["contrast_factor",       "barns",   10.0,  [-inf, inf], "", "Contrast factor of the polymer"],
    ["bjerrum_length",        "Ang",      7.1,  [0, inf],    "", "Bjerrum length"],
    ["virial_param",          "Ang^3", 12.0,  [-inf, inf], "", "Virial parameter"],
    ["monomer_length",        "Ang",     10.0,  [0, inf],    "", "Monomer length"],
    ["salt_concentration",    "mol/L",    0.0,  [-inf, inf], "", "Concentration of monovalent salt"],
    ["ionization_degree",     "",         0.05, [0, inf],    "", "Degree of ionization"],
    ["polymer_concentration", "mol/L",    0.7,  [0, inf],    "", "Polymer molar concentration"],
    ]
# pylint: enable=bad-whitespace, line-too-long


def Iq(q,
       contrast_factor,
       bjerrum_length,
       virial_param,
       monomer_length,
       salt_concentration,
       ionization_degree,
       polymer_concentration):
    """
    :params: see parameter table
    :return: 1-D form factor for polyelectrolytes in low salt

    parameter names, units, default values, and behavior (volume, sld etc) are
    defined in the parameter table.  The concentrations are converted from
    experimental mol/L to dimensionally useful 1/A3 in first two lines
    """

    concentration_pol = polymer_concentration * 6.022136e-4
    concentration_salt = salt_concentration * 6.022136e-4

    k_square = 4.0 * pi * bjerrum_length * (2*concentration_salt +
                                            ionization_degree * concentration_pol)

    r0_square = 1.0/ionization_degree/sqrt(concentration_pol) * \
                (monomer_length/sqrt((48.0*pi*bjerrum_length)))

    term1 = contrast_factor/(4.0 * pi * bjerrum_length *
                             ionization_degree**2) * (q**2 + k_square)

    term2 = 1.0 + r0_square**2 * (q**2 + k_square) * \
        (q**2 - (12.0 * virial_param * concentration_pol/(monomer_length**2)))

    return term1/term2

Iq.vectorized = True  # Iq accepts an array of q values

def random():
    """Return a random parameter set for the model."""
    # TODO: review random be_polyelectrolyte model generation
    pars = dict(
        scale=10000, #background=0,
        #polymer_concentration=0.7,
        polymer_concentration=np.random.beta(5, 3), # around 70%
        #salt_concentration=0.0,
        # keep salt concentration extremely low
        # and use explicit molar to match polymer concentration
        salt_concentration=np.random.beta(1, 100)*6.022136e-4,
        #contrast_factor=10.0,
        contrast_fact=np.random.uniform(1, 100),
        #bjerrum_length=7.1,
        bjerrum_length=np.random.uniform(1, 10),
        #virial_param=12.0,
        virial_param=np.random.uniform(-1000, 30),
        #monomer_length=10.0,
        monomer_length=10.0**(4*np.random.beta(1.5, 3)),
        #ionization_degree=0.05,
        ionization_degree=np.random.beta(1.5, 4),
    )
    return pars

tests = [

    # Accuracy tests based on content in test/utest_other_models.py
    # Note that these should some day be validated beyond this self validation
    # (circular reasoning). -- i.e. the "good value," at least for those with
    # non zero salt concentrations, were obtained by running the current
    # model in SasView and copying the appropriate result here.
    #    PDB -- Sep 26, 2018
    [{'contrast_factor':       10.0,
      'bjerrum_length':         7.1,
      'virial_param':          12.0,
      'monomer_length':        10.0,
      'salt_concentration':     0.0,
      'ionization_degree':      0.05,
      'polymer_concentration':  0.7,
      'background':             0.001,
     }, 0.001, 0.0948379],

    [{'contrast_factor':       10.0,
      'bjerrum_length':       100.0,
      'virial_param':           3.0,
      'monomer_length':         5.0,
      'salt_concentration':     1.0,
      'ionization_degree':      0.1,
      'polymer_concentration':  1.0,
      'background':             0.0,
     }, 0.1, 0.253469484],

    [{'contrast_factor':       10.0,
      'bjerrum_length':       100.0,
      'virial_param':           3.0,
      'monomer_length':         5.0,
      'salt_concentration':     1.0,
      'ionization_degree':      0.1,
      'polymer_concentration':  1.0,
      'background':             1.0,
     }, 0.05, 1.738358122],

    [{'contrast_factor':     100.0,
      'bjerrum_length':       10.0,
      'virial_param':         12.0,
      'monomer_length':       10.0,
      'salt_concentration':    0.1,
      'ionization_degree':     0.5,
      'polymer_concentration': 0.1,
      'background':           0.01,
     }, 0.5, 0.012881893],
    ]



from __future__ import division, print_function
# Make sasmodels available on the path
import sys, os
BETA_DIR = os.path.dirname(os.path.realpath(__file__))
SASMODELS_DIR = os.path.dirname(os.path.dirname(BETA_DIR))
sys.path.insert(0, SASMODELS_DIR)

from collections import namedtuple

from matplotlib import pyplot as plt
import numpy as np
from numpy import pi, sin, cos, sqrt, fabs
from numpy.polynomial.legendre import leggauss
from scipy.special import j1 as J1
from numpy import inf
from scipy.special import gammaln  # type: ignore

Theory = namedtuple('Theory', 'Q F1 F2 P S I Seff Ibeta')
Theory.__new__.__defaults__ = (None,) * len(Theory._fields)

#Used to calculate F(q) for the cylinder, sphere, ellipsoid models
def sas_sinx_x(x):
    with np.errstate(all='ignore'):
        retvalue = sin(x)/x
    retvalue[x == 0.] = 1.
    return retvalue

def sas_2J1x_x(x):
    with np.errstate(all='ignore'):
        retvalue = 2*J1(x)/x
    retvalue[x == 0] = 1.
    return retvalue

def sas_3j1x_x(x):
    """return 3*j1(x)/x"""
    retvalue = np.empty_like(x)
    with np.errstate(all='ignore'):
        # GSL bessel_j1 taylor expansion
        index = (x < 0.25)
        y = x[index]**2
        c1 = -1.0/10.0
        c2 = +1.0/280.0
        c3 = -1.0/15120.0
        c4 = +1.0/1330560.0
        c5 = -1.0/172972800.0
        retvalue[index] = 1.0 + y*(c1 + y*(c2 + y*(c3 + y*(c4 + y*c5))))
        index = ~index
        y = x[index]
        retvalue[index] = 3*(sin(y) - y*cos(y))/y**3
    retvalue[x == 0.] = 1.
    return retvalue

#Used to cross check my models with sasview models
def build_model(model_name, q, **pars):
    from sasmodels.core import load_model_info, build_model as build_sasmodel
    from sasmodels.data import empty_data1D
    from sasmodels.direct_model import DirectModel
    model_info = load_model_info(model_name)
    model = build_sasmodel(model_info, dtype='double!')
    data = empty_data1D(q)
    calculator = DirectModel(data, model,cutoff=0)
    calculator.pars = pars.copy()
    calculator.pars.setdefault('background', 0)
    return calculator

#gives the hardsphere structure factor that sasview uses
def _hardsphere_simple(q, radius_effective, volfraction):
    CUTOFFHS = 0.05
    if fabs(radius_effective) < 1.E-12:
        HARDSPH = 1.0
        return HARDSPH
    X = 1.0/(1.0 -volfraction)
    D = X*X
    A = (1.+2.*volfraction)*D
    A *= A
    X = fabs(q*radius_effective*2.0)
    if X < 5.E-06:
        HARDSPH = 1./A
        return HARDSPH
    X2 = X*X
    B = (1.0 +0.5*volfraction)*D
    B *= B
    B *= -6.*volfraction
    G = 0.5*volfraction*A
    if X < CUTOFFHS:
        FF = 8.0*A +6.0*B + 4.0*G + ( -0.8*A -B/1.5 -0.5*G +(A/35. +0.0125*B +0.02*G)*X2)*X2
        HARDSPH = 1./(1. + volfraction*FF )
        return HARDSPH
    X4 = X2*X2
    S, C = sin(X), cos(X)
    FF =  ((G*( (4.*X2 -24.)*X*S -(X4 -12.*X2 +24.)*C +24. )/X2 + B*(2.*X*S -(X2-2.)*C -2.) )/X + A*(S-X*C))/X
    HARDSPH = 1./(1. + 24.*volfraction*FF/X2)
    return HARDSPH

def hardsphere_simple(q, radius_effective, volfraction):
    SQ = [_hardsphere_simple(qk, radius_effective, volfraction) for qk in q]
    return np.array(SQ)

#Used in gaussian quadrature for polydispersity
#returns values and the probability of those values based on gaussian distribution
N_GAUSS = 35
NSIGMA_GAUSS = 3
def gaussian_distribution(center, sigma, lb, ub):
    #3 standard deviations covers approx. 99.7%
    if sigma != 0:
        nsigmas = NSIGMA_GAUSS
        x = np.linspace(center-sigma*nsigmas, center+sigma*nsigmas, num=N_GAUSS)
        x = x[(x >= lb) & (x <= ub)]
        px = np.exp((x-center)**2 / (-2.0 * sigma * sigma))
        return x, px
    else:
        return np.array([center]), np.array([1])

N_SCHULZ = 80
NSIGMA_SCHULZ = 8
def schulz_distribution(center, sigma, lb, ub):
    if sigma != 0:
        nsigmas = NSIGMA_SCHULZ
        x = np.linspace(center-sigma*nsigmas, center+sigma*nsigmas, num=N_SCHULZ)
        x = x[(x >= lb) & (x <= ub)]
        R = x/center
        z = (center/sigma)**2
        arg = z*np.log(z) + (z-1)*np.log(R) - R*z - np.log(center) - gammaln(z)
        px = np.exp(arg)
        return x, px
    else:
        return np.array([center]), np.array([1])

#returns the effective radius used in sasview
def ER_ellipsoid(radius_polar, radius_equatorial):
    ee = np.empty_like(radius_polar)
    if radius_polar > radius_equatorial:
        ee = (radius_polar**2 - radius_equatorial**2)/radius_polar**2
    elif radius_polar < radius_equatorial:
        ee = (radius_equatorial**2 - radius_polar**2) / radius_equatorial**2
    else:
        ee = 2*radius_polar
    if radius_polar * radius_equatorial != 0:
        bd = 1.0 - ee
        e1 = np.sqrt(ee)
        b1 = 1.0 + np.arcsin(e1) / (e1*np.sqrt(bd))
        bL = (1.0 + e1) / (1.0 - e1)
        b2 = 1.0 + bd / 2 / e1 * np.log(bL)
        delta = 0.75 * b1 * b2
    ddd = np.zeros_like(radius_polar)
    ddd = 2.0*(delta + 1.0)*radius_polar*radius_equatorial**2
    return 0.5*ddd**(1.0 / 3.0)

def ellipsoid_volume(radius_polar, radius_equatorial):
    volume = (4./3.)*pi*radius_polar*radius_equatorial**2
    return volume

# F1 is F(q)
# F2 is F(g)^2
#IQM is I(q) with monodispersity
#IQSM is I(q) with structure factor S(q) and monodispersity
#IQBM is I(q) with Beta Approximation and monodispersity
#SQ is monodisperse approach for structure factor
#SQ_EFF is the effective structure factor from beta approx
def ellipsoid_theta(q, radius_polar, radius_equatorial, sld, sld_solvent,
                    volfraction=0, radius_effective=None):
    #creates values z and corresponding probabilities w from legendre-gauss quadrature
    volume = ellipsoid_volume(radius_polar, radius_equatorial)
    z, w = leggauss(76)
    F1 = np.zeros_like(q)
    F2 = np.zeros_like(q)
    #use a u subsition(u=cos) and then u=(z+1)/2 to change integration from
    #0->2pi with respect to alpha to -1->1 with respect to z, allowing us to use
    #legendre-gauss quadrature
    for k, qk in enumerate(q):
        r = sqrt(radius_equatorial**2*(1-((z+1)/2)**2)+radius_polar**2*((z+1)/2)**2)
        form = (sld-sld_solvent)*volume*sas_3j1x_x(qk*r)
        F2[k] = np.sum(w*form**2)
        F1[k] = np.sum(w*form)
    #the 1/2 comes from the change of variables mentioned above
    F2 = F2/2.0
    F1 = F1/2.0
    if radius_effective is None:
        radius_effective = ER_ellipsoid(radius_polar,radius_equatorial)
    SQ = hardsphere_simple(q, radius_effective, volfraction)
    SQ_EFF = 1 + F1**2/F2*(SQ - 1)
    IQM = 1e-4*F2/volume
    IQSM = volfraction*IQM*SQ
    IQBM = volfraction*IQM*SQ_EFF
    return Theory(Q=q, F1=F1, F2=F2, P=IQM, S=SQ, I=IQSM, Seff=SQ_EFF, Ibeta=IQBM)

#IQD is I(q) polydispursed, IQSD is I(q)S(q) polydispursed, etc.
#IQBD HAS NOT BEEN CROSS CHECKED AT ALL
def ellipsoid_pe(q, radius_polar, radius_equatorial, sld, sld_solvent,
                 radius_polar_pd=0.1, radius_equatorial_pd=0.1,
                 radius_polar_pd_type='gaussian',
                 radius_equatorial_pd_type='gaussian',
                 volfraction=0, radius_effective=None,
                 background=0, scale=1,
                 norm='sasview'):
    if norm not in ['sasview', 'sasfit', 'yun']:
        raise TypeError("unknown norm "+norm)
    if radius_polar_pd_type == 'gaussian':
        Rp_val, Rp_prob = gaussian_distribution(radius_polar, radius_polar_pd*radius_polar, 0, inf)
    elif radius_polar_pd_type == 'schulz':
        Rp_val, Rp_prob = schulz_distribution(radius_polar, radius_polar_pd*radius_polar, 0, inf)
    if radius_equatorial_pd_type == 'gaussian':
        Re_val, Re_prob = gaussian_distribution(radius_equatorial, radius_equatorial_pd*radius_equatorial, 0, inf)
    elif radius_equatorial_pd_type == 'schulz':
        Re_val, Re_prob = schulz_distribution(radius_equatorial, radius_equatorial_pd*radius_equatorial, 0, inf)
    total_weight = total_volume = 0
    radius_eff = 0
    F1, F2 = np.zeros_like(q), np.zeros_like(q)
    for k, Rpk in enumerate(Rp_val):
        print("ellipsoid cycle", k, "of", len(Rp_val))
        for i, Rei in enumerate(Re_val):
            theory = ellipsoid_theta(q, Rpk, Rei, sld, sld_solvent)
            volume = ellipsoid_volume(Rpk, Rei)
            weight = Rp_prob[k]*Re_prob[i]
            total_weight += weight
            total_volume += weight*volume
            F1 += theory.F1*weight
            F2 += theory.F2*weight
            radius_eff += weight*ER_ellipsoid(Rpk, Rei)
    F1 /= total_weight
    F2 /= total_weight
    average_volume = total_volume/total_weight
    if radius_effective is None:
        radius_effective = radius_eff/total_weight
    if norm == 'sasfit':
        IQD = F2
    elif norm == 'sasview':
        # Note: internally, sasview uses F2/total_volume because:
        #   average_volume = total_volume/total_weight
        #   F2/total_weight / average_volume
        #     = F2/total_weight / total_volume/total_weight
        #     = F2/total_volume
        IQD = F2/average_volume*1e-4*volfraction
        F1 *= 1e-2  # Yun is using sld in 1/A^2, not 1e-6/A^2
        F2 *= 1e-4
    elif norm == 'yun':
        F1 *= 1e-6  # Yun is using sld in 1/A^2, not 1e-6/A^2
        F2 *= 1e-12
        IQD = F2/average_volume*1e8*volfraction
    SQ = hardsphere_simple(q, radius_effective, volfraction)
    beta = F1**2/F2
    SQ_EFF = 1 + beta*(SQ - 1)
    IQSD = IQD*SQ
    IQBD = IQD*SQ_EFF
    return Theory(Q=q, F1=F1, F2=F2, P=IQD, S=SQ, I=IQSD, Seff=SQ_EFF, Ibeta=IQBD)

#polydispersity for sphere
def sphere_r(q,radius,sld,sld_solvent,
             radius_pd=0.1, radius_pd_type='gaussian',
             volfraction=0, radius_effective=None,
             background=0, scale=1,
             norm='sasview'):
    if norm not in ['sasview', 'sasfit', 'yun']:
        raise TypeError("unknown norm "+norm)
    if radius_pd_type == 'gaussian':
        radius_val, radius_prob = gaussian_distribution(radius, radius_pd*radius, 0, inf)
    elif radius_pd_type == 'schulz':
        radius_val, radius_prob = schulz_distribution(radius, radius_pd*radius, 0, inf)
    total_weight = total_volume = 0
    F1 = np.zeros_like(q)
    F2 = np.zeros_like(q)
    for k, rk in enumerate(radius_val):
        volume = 4./3.*pi*rk**3
        total_weight += radius_prob[k]
        total_volume += radius_prob[k]*volume
        form = (sld-sld_solvent)*volume*sas_3j1x_x(q*rk)
        F2 += radius_prob[k]*form**2
        F1 += radius_prob[k]*form
    F1 /= total_weight
    F2 /= total_weight
    average_volume = total_volume/total_weight

    if radius_effective is None:
        radius_effective = radius
    average_volume = total_volume/total_weight
    if norm == 'sasfit':
        IQD = F2
    elif norm == 'sasview':
        IQD = F2/average_volume*1e-4*volfraction
    elif norm == 'yun':
        F1 *= 1e-6  # Yun is using sld in 1/A^2, not 1e-6/A^2
        F2 *= 1e-12
        IQD = F2/average_volume*1e8*volfraction
    SQ = hardsphere_simple(q, radius_effective, volfraction)
    beta = F1**2/F2
    SQ_EFF = 1 + beta*(SQ - 1)
    IQSD = IQD*SQ
    IQBD = IQD*SQ_EFF
    return Theory(Q=q, F1=F1, F2=F2, P=IQD, S=SQ, I=IQSD, Seff=SQ_EFF, Ibeta=IQBD)

###############################################################################
###############################################################################
###############################################################################
##################                                           ##################
##################                   TESTS                   ##################
##################                                           ##################
###############################################################################
###############################################################################
###############################################################################

def popn(d, keys):
    """
    Splits a dict into two, with any key of *d* which is in *keys* removed
    from *d* and added to *b*. Returns *b*.
    """
    b = {}
    for k in keys:
        try:
            b[k] = d.pop(k)
        except KeyError:
            pass
    return b

def sasmodels_theory(q, Pname, **pars):
    """
    Call sasmodels to compute the model with and without a hard sphere
    structure factor.
    """
    #mono_pars = {k: (0 if k.endswith('_pd') else v) for k, v in pars.items()}
    Ppars = pars.copy()
    Spars = popn(Ppars, ['radius_effective', 'volfraction'])
    Ipars = pars.copy()

    # Autofill npts and nsigmas for the given pd type
    for k, v in pars.items():
        if k.endswith("_pd_type"):
            if v == "gaussian":
                n, nsigmas = N_GAUSS, NSIGMA_GAUSS
            elif v == "schulz":
                n, nsigmas = N_SCHULZ, NSIGMA_SCHULZ
            Ppars.setdefault(k.replace("_pd_type", "_pd_n"), n)
            Ppars.setdefault(k.replace("_pd_type", "_pd_nsigma"), nsigmas)
            Ipars.setdefault(k.replace("_pd_type", "_pd_n"), n)
            Ipars.setdefault(k.replace("_pd_type", "_pd_nsigma"), nsigmas)

    #Ppars['scale'] = Spars.get('volfraction', 1)
    P = build_model(Pname, q)
    S = build_model("hardsphere", q)
    I = build_model(Pname+"@hardsphere", q)
    Pq = P(**Ppars)*pars.get('volfraction', 1)
    Sq = S(**Spars)
    Iq = I(**Ipars)
    #Iq = Pq*Sq*pars.get('volfraction', 1)
    #Sq = Iq/Pq
    #Iq = None#= Sq = None
    r = dict(I._kernel.results())
    return Theory(Q=q, F1=None, F2=None, P=Pq, S=None, I=None, Seff=r["S_eff(Q)"], Ibeta=Iq)

def compare(title, target, actual, fields='F1 F2 P S I Seff Ibeta'):
    """
    Plot fields in common between target and actual, along with relative error.
    """
    available = [s for s in fields.split()
                 if getattr(target, s) is not None and getattr(actual, s) is not None]
    rows = len(available)
    for row, field in enumerate(available):
        Q = target.Q
        I1, I2 = getattr(target, field), getattr(actual, field)
        plt.subplot(rows, 2, 2*row+1)
        plt.loglog(Q, abs(I1), label="target "+field)
        plt.loglog(Q, abs(I2), label="value "+field)
        #plt.legend(loc="upper left", bbox_to_anchor=(1,1))
        plt.legend(loc='lower left')
        plt.subplot(rows, 2, 2*row+2)
        plt.semilogx(Q, I2/I1 - 1, label="relative error")
        #plt.semilogx(Q, I1/I2 - 1, label="relative error")
    plt.tight_layout()
    plt.suptitle(title)
    plt.show()

def data_file(name):
    return os.path.join(BETA_DIR, 'data_files', name)

def load_sasfit(path):
    data = np.loadtxt(path, dtype=str, delimiter=';').T
    data = np.vstack((map(float, v) for v in data[0:2]))
    return data

COMPARISON = {}  # Type: Dict[(str,str,str)] -> Callable[(), None]

def compare_sasview_sphere(pd_type='schulz'):
    q = np.logspace(-5, 0, 250)
    model = 'sphere'
    pars = dict(
        radius=20, sld=4, sld_solvent=1,
        background=0,
        radius_pd=.1, radius_pd_type=pd_type,
        volfraction=0.15,
        #radius_effective=12.59921049894873,  # equivalent average sphere radius
        )
    target = sasmodels_theory(q, model, **pars)
    actual = sphere_r(q, norm='sasview', **pars)
    title = " ".join(("sasmodels", model, pd_type))
    compare(title, target, actual)
COMPARISON[('sasview', 'sphere', 'gaussian')] = lambda: compare_sasview_sphere(pd_type='gaussian')
COMPARISON[('sasview', 'sphere', 'schulz')] = lambda: compare_sasview_sphere(pd_type='schulz')

def compare_sasview_ellipsoid(pd_type='gaussian'):
    q = np.logspace(-5, 0, 50)
    model = 'ellipsoid'
    pars = dict(
        radius_polar=20, radius_equatorial=400, sld=4, sld_solvent=1,
        background=0,
        radius_polar_pd=0.1, radius_polar_pd_type=pd_type,
        radius_equatorial_pd=0.1, radius_equatorial_pd_type=pd_type,
        volfraction=0.15,
        radius_effective=270.7543927018,
        #radius_effective=12.59921049894873,
        )
    target = sasmodels_theory(q, model, radius_effective_mode=0, structure_factor_mode=1, **pars)
    actual = ellipsoid_pe(q, norm='sasview', **pars)
    title = " ".join(("sasmodels", model, pd_type))
    compare(title, target, actual)
COMPARISON[('sasview', 'ellipsoid', 'gaussian')] = lambda: compare_sasview_ellipsoid(pd_type='gaussian')
COMPARISON[('sasview', 'ellipsoid', 'schulz')] = lambda: compare_sasview_ellipsoid(pd_type='schulz')

def compare_yun_ellipsoid_mono():
    pars = {
        'radius_polar': 20, 'radius_polar_pd': 0, 'radius_polar_pd_type': 'gaussian',
        'radius_equatorial': 10, 'radius_equatorial_pd': 0, 'radius_equatorial_pd_type': 'gaussian',
        'sld': 2, 'sld_solvent': 1,
        'volfraction': 0.15,
        # Yun uses radius for same volume sphere for effective radius
        # whereas sasview uses the average curvature.
        'radius_effective': 12.59921049894873,
    }

    data = np.loadtxt(data_file('yun_ellipsoid.dat'),skiprows=2).T
    Q = data[0]
    F1 = data[1]
    P = data[3]*pars['volfraction']
    S = data[5]
    Seff = data[6]
    target = Theory(Q=Q, F1=F1, P=P, S=S, Seff=Seff)
    actual = ellipsoid_pe(Q, norm='yun', **pars)
    title = " ".join(("yun", "ellipsoid", "no pd"))
    #compare(title, target, actual, fields="P S I Seff Ibeta")
    compare(title, target, actual)
COMPARISON[('yun', 'ellipsoid', 'gaussian')] = compare_yun_ellipsoid_mono
COMPARISON[('yun', 'ellipsoid', 'schulz')] = compare_yun_ellipsoid_mono

def compare_yun_sphere_gauss():
    # Note: yun uses gauss limits from R0/10 to R0 + 5 sigma steps sigma/100
    # With pd = 0.1, that's 14 sigma and 1400 points.
    pars = {
        'radius': 20, 'radius_pd': 0.1, 'radius_pd_type': 'gaussian',
        'sld': 6, 'sld_solvent': 0,
        'volfraction': 0.1,
    }

    data = np.loadtxt(data_file('testPolydisperseGaussianSphere.dat'), skiprows=2).T
    Q = data[0]
    F1 = data[1]
    F2 = data[2]
    P = data[3]
    S = data[5]
    Seff = data[6]
    target = Theory(Q=Q, F1=F1, P=P, S=S, Seff=Seff)
    actual = sphere_r(Q, norm='yun', **pars)
    title = " ".join(("yun", "sphere", "10% dispersion 10% Vf"))
    compare(title, target, actual)
    data = np.loadtxt(data_file('testPolydisperseGaussianSphere2.dat'), skiprows=2).T
    pars.update(radius_pd=0.15)
    Q = data[0]
    F1 = data[1]
    F2 = data[2]
    P = data[3]
    S = data[5]
    Seff = data[6]
    target = Theory(Q=Q, F1=F1, P=P, S=S, Seff=Seff)
    actual = sphere_r(Q, norm='yun', **pars)
    title = " ".join(("yun", "sphere", "15% dispersion 10% Vf"))
    compare(title, target, actual)
COMPARISON[('yun', 'sphere', 'gaussian')] = compare_yun_sphere_gauss


def compare_sasfit_sphere_gauss():
    #N=1,s=2,X0=20,distr radius R=20,eta_core=4,eta_solv=1,.3
    pars = {
        'radius': 20, 'radius_pd': 0.1, 'radius_pd_type': 'gaussian',
        'sld': 4, 'sld_solvent': 1,
        'volfraction': 0.3,
    }

    Q, IQ = load_sasfit(data_file('sasfit_sphere_IQD.txt'))
    Q, IQSD = load_sasfit(data_file('sasfit_sphere_IQSD.txt'))
    Q, IQBD = load_sasfit(data_file('sasfit_sphere_IQBD.txt'))
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_sphere_sq.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_sphere_sqeff.txt'))
    target = Theory(Q=Q, F1=None, F2=None, P=IQ, S=SQ, I=IQSD, Seff=SQ_EFF, Ibeta=IQBD)
    actual = sphere_r(Q, norm="sasfit", **pars)
    title = " ".join(("sasfit", "sphere", "pd=10% gaussian"))
    compare(title, target, actual)
    #compare(title, target, actual, fields="P")
COMPARISON[('sasfit', 'sphere', 'gaussian')] = compare_sasfit_sphere_gauss

def compare_sasfit_sphere_schulz():
    #radius=20,sld=4,sld_solvent=1,volfraction=0.3,radius_pd=0.1
    #We have scaled the output from sasfit by 1e-4*volume*volfraction
    #0.10050378152592121
    pars = {
        'radius': 20, 'radius_pd': 0.1, 'radius_pd_type': 'schulz',
        'sld': 4, 'sld_solvent': 1,
        'volfraction': 0.3,
    }

    Q, IQ = load_sasfit(data_file('sasfit_sphere_schulz_IQD.txt'))
    Q, IQSD = load_sasfit(data_file('sasfit_sphere_schulz_IQSD.txt'))
    Q, IQBD = load_sasfit(data_file('sasfit_sphere_schulz_IQBD.txt'))
    target = Theory(Q=Q, F1=None, F2=None, P=IQ, S=None, I=IQSD, Seff=None, Ibeta=IQBD)
    actual = sphere_r(Q, norm="sasfit", **pars)
    title = " ".join(("sasfit", "sphere", "pd=10% schulz"))
    compare(title, target, actual)
COMPARISON[('sasfit', 'sphere', 'schulz')] = compare_sasfit_sphere_schulz

def compare_sasfit_ellipsoid_schulz():
    #polarradius=20, equatorialradius=10, sld=4,sld_solvent=1,volfraction=0.3,radius_polar_pd=0.1
    #Effective radius =13.1353356684
    #We have scaled the output from sasfit by 1e-4*volume*volfraction
    #0.10050378152592121
    pars = {
        'radius_polar': 20, 'radius_polar_pd': 0.1, 'radius_polar_pd_type': 'schulz',
        'radius_equatorial': 10, 'radius_equatorial_pd': 0., 'radius_equatorial_pd_type': 'schulz',
        'sld': 4, 'sld_solvent': 1,
        'volfraction': 0.3, 'radius_effective': 13.1353356684,
    }

    Q, IQ = load_sasfit(data_file('sasfit_ellipsoid_shulz_IQD.txt'))
    Q, IQSD = load_sasfit(data_file('sasfit_ellipsoid_shulz_IQSD.txt'))
    Q, IQBD = load_sasfit(data_file('sasfit_ellipsoid_shulz_IQBD.txt'))
    target = Theory(Q=Q, F1=None, F2=None, P=IQ, S=None, I=IQSD, Seff=None, Ibeta=IQBD)
    actual = ellipsoid_pe(Q, norm="sasfit", **pars)
    title = " ".join(("sasfit", "ellipsoid", "pd=10% schulz"))
    compare(title, target, actual)
COMPARISON[('sasfit', 'ellipsoid', 'schulz')] = compare_sasfit_ellipsoid_schulz


def compare_sasfit_ellipsoid_gaussian():
    pars = {
        'radius_polar': 20, 'radius_polar_pd': 0, 'radius_polar_pd_type': 'gaussian',
        'radius_equatorial': 10, 'radius_equatorial_pd': 0, 'radius_equatorial_pd_type': 'gaussian',
        'sld': 4, 'sld_solvent': 1,
        'volfraction': 0, 'radius_effective': None,
    }

    #Rp=20,Re=10,eta_core=4,eta_solv=1
    Q, PQ0 = load_sasfit(data_file('sasfit_ellipsoid_IQM.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.0, radius_equatorial_pd=0, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ0)
    compare("sasfit ellipsoid no poly", target, actual); plt.show()

    #N=1,s=2,X0=20,distr 10% polar Rp=20,Re=10,eta_core=4,eta_solv=1, no structure poly
    Q, PQ_Rp10 = load_sasfit(data_file('sasfit_ellipsoid_IQD.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.1, radius_equatorial_pd=0.0, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ_Rp10)
    compare("sasfit ellipsoid P(Q) 10% Rp", target, actual); plt.show()
    #N=1,s=1,X0=10,distr 10% equatorial Rp=20,Re=10,eta_core=4,eta_solv=1, no structure poly
    Q, PQ_Re10 = load_sasfit(data_file('sasfit_ellipsoid_IQD2.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.0, radius_equatorial_pd=0.1, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ_Re10)
    compare("sasfit ellipsoid P(Q) 10% Re", target, actual); plt.show()
    #N=1,s=6,X0=20,distr 30% polar Rp=20,Re=10,eta_core=4,eta_solv=1, no structure poly
    Q, PQ_Rp30 = load_sasfit(data_file('sasfit_ellipsoid_IQD3.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.3, radius_equatorial_pd=0.0, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ_Rp30)
    compare("sasfit ellipsoid P(Q) 30% Rp", target, actual); plt.show()
    #N=1,s=3,X0=10,distr 30% equatorial Rp=20,Re=10,eta_core=4,eta_solv=1, no structure poly
    Q, PQ_Re30 = load_sasfit(data_file('sasfit_ellipsoid_IQD4.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.0, radius_equatorial_pd=0.3, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ_Re30)
    compare("sasfit ellipsoid P(Q) 30% Re", target, actual); plt.show()
    #N=1,s=12,X0=20,distr 60% polar Rp=20,Re=10,eta_core=4,eta_solv=1, no structure poly
    Q, PQ_Rp60 = load_sasfit(data_file('sasfit_ellipsoid_IQD5.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.6, radius_equatorial_pd=0.0, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ_Rp60)
    compare("sasfit ellipsoid P(Q) 60% Rp", target, actual); plt.show()
    #N=1,s=6,X0=10,distr 60% equatorial Rp=20,Re=10,eta_core=4,eta_solv=1, no structure poly
    Q, PQ_Re60 = load_sasfit(data_file('sasfit_ellipsoid_IQD6.txt'))
    pars.update(volfraction=0, radius_polar_pd=0.0, radius_equatorial_pd=0.6, radius_effective=None)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, P=PQ_Re60)
    compare("sasfit ellipsoid P(Q) 60% Re", target, actual); plt.show()

    #N=1,s=2,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.1354236254,.15
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff.txt'))
    pars.update(volfraction=0.15, radius_polar_pd=0.1, radius_equatorial_pd=0, radius_effective=13.1354236254)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 10% Rp 15% Vf", target, actual); plt.show()
    #N=1,s=6,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.0901197149,.15
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq2.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff2.txt'))
    pars.update(volfraction=0.15, radius_polar_pd=0.3, radius_equatorial_pd=0, radius_effective=13.0901197149)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 30% Rp 15% Vf", target, actual); plt.show()
    #N=1,s=12,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.336060917,.15
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq3.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff3.txt'))
    pars.update(volfraction=0.15, radius_polar_pd=0.6, radius_equatorial_pd=0, radius_effective=13.336060917)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 60% Rp 15% Vf", target, actual); plt.show()

    #N=1,s=2,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.1354236254,.3
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq4.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff4.txt'))
    pars.update(volfraction=0.3, radius_polar_pd=0.1, radius_equatorial_pd=0, radius_effective=13.1354236254)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 10% Rp 30% Vf", target, actual); plt.show()
    #N=1,s=6,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.0901197149,.3
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq5.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff5.txt'))
    pars.update(volfraction=0.3, radius_polar_pd=0.3, radius_equatorial_pd=0, radius_effective=13.0901197149)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 30% Rp 30% Vf", target, actual); plt.show()
    #N=1,s=12,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.336060917,.3
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq6.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff6.txt'))
    pars.update(volfraction=0.3, radius_polar_pd=0.6, radius_equatorial_pd=0, radius_effective=13.336060917)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 60% Rp 30% Vf", target, actual); plt.show()

    #N=1,s=2,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.1354236254,.6
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq7.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff7.txt'))
    pars.update(volfraction=0.6, radius_polar_pd=0.1, radius_equatorial_pd=0, radius_effective=13.1354236254)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 10% Rp 60% Vf", target, actual); plt.show()
    #N=1,s=6,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.0901197149,.6
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq8.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff8.txt'))
    pars.update(volfraction=0.6, radius_polar_pd=0.3, radius_equatorial_pd=0, radius_effective=13.0901197149)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 30% Rp 60% Vf", target, actual); plt.show()
    #N=1,s=12,X0=20,distr polar Rp=20,Re=10,eta_core=4,eta_solv=1, hardsphere ,13.336060917,.6
    Q, SQ = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sq9.txt'))
    Q, SQ_EFF = load_sasfit(data_file('sasfit_polydisperse_ellipsoid_sqeff9.txt'))
    pars.update(volfraction=0.6, radius_polar_pd=0.6, radius_equatorial_pd=0, radius_effective=13.336060917)
    actual = ellipsoid_pe(Q, norm='sasfit', **pars)
    target = Theory(Q=Q, S=SQ, Seff=SQ_EFF)
    compare("sasfit ellipsoid P(Q) 60% Rp 60% Vf", target, actual); plt.show()
COMPARISON[('sasfit', 'ellipsoid', 'gaussian')] = compare_sasfit_ellipsoid_gaussian

def main():
    key = tuple(sys.argv[1:])
    if key not in COMPARISON:
        print("usage: sasfit_compare.py [sasview|sasfit|yun] [sphere|ellipsoid] [gaussian|schulz]")
        return
    comparison = COMPARISON[key]
    comparison()

if __name__ == "__main__":
    main()



r"""
This model is a special case of the power law model, deriving its special name
from the Porod Law which it models, where the power law exponent is fixed to -4.

.. math:: I(q) = C/q^4 + Background

Where the power law constant $C$ in this case is just the scale factor.

The Porod law, similar to the Guinier model, is a broadly applicable model to a
very restricted portion of the data. While the Guinier model applies to any
dilute particulate system regardless of shape or size but only to the very low
$q$ (being a taylor expansion around $q$ = 0), the Porod Law applies to any
scattering system with sharp scattering length density interfaces between the
phases but only at very high $q$ (technically in the limit as $q \to \infty$ ..
except of course that, practically, the continuum approach breaks down there).
It is based on the idea that at sufficiently high $q$ there is no shape
information left and all scattering is just reflections off the sharp
interfaces.

In the special case of a two phase system, the power law constant $C$ derived
from the appropriate $q$ limit portion of the data is known as the Porod
Constant and can be written as:

.. math:: C = 2\pi (\Delta\rho)^2 S_v

where $S_v$ is the specific surface area (ie, surface area / volume) of the
material under study, and $\Delta\rho$ is the contrast factor between the two
phases.

Thus, by extracting the Porod Constant from experimental data, and
knowing the contrast factor between the two phases, one can obtain the specific
surface area for the material. This can be very useful for example in
understanding porosity in materials, particularly when used in conjunction
with complementary techniques such as BET.

.. Note:: While it is straightforward to calculate the specific surface area,
    $S_v$ of the material from the Porod Constant returned here, the Invariant
    analysis panel will also compute the $S_v$ by entering the contrast term
    and the Porod Constant obtained here into the appropriate entry boxes.

    There are, however, several caveats regarding obtaining a good experimental
    value of the Porod Constant.

    * First, as it is the scale value, the data **must** be on an absolute scale.

    * Next of course, there must be a sufficiently large $q$ range that is in
      the Porod region to be able to be fit. But this may not always be
      possible: for example, polymer coils in solution will often not reach that
      limit within typical SAS ranges, nor even the $q$ ranges where the
      continuum approach of using the SLD is even valid.

    * For highly monodisperse systems with limited resolution smearing, the
      data will contain large oscillations which will make the estimate from
      this fit unreliable.  It will vary depending on the exact range of the fit
      that is chosen. This is because numerical integration using a finite
      number of points cannot properly capture the exact area under the data
      across these very sharp dips.

    * Ironically, large resolution smearing, and/or polydispersity smearing,
      will make the value obtained much more consistent and reliable. Thus the
      problem is less severe for typical real data than for simulated data that
      does not simulate any resolution smearing.

    * If in any doubt, using several values of $q_{min}$ in the fitting, and
      comparing with the value obtained in a different manner, is highly
      advisable. One such method uses the Porod plot ($I(q)* q^4$ vs. $q^4$).
      Fitting the highest $q$ (strictly speaking $q^4$) data to a straight line
      using as much of the high $q$ as fits well to a straight line, yields a
      slope which is the flat background and an intercept which is the Porod
      Constant.

For 2D data: The 2D scattering intensity is calculated in the same way as 1D,
where the q vector is defined as

.. math:: q = \sqrt{q_x^2+q_y^2}

References
----------

#. G Porod. *Kolloid Zeit*. 124 (1951) 83
#. L A Feigin, D I Svergun, G W Taylor. *Structure Analysis by Small-Angle
   X-ray and Neutron Scattering*. Springer. (1987)

Authorship and Verification
----------------------------

* **Author:**
* **Last Modified by: Paul Butler, 28Mar2020**
* **Last Reviewed by:**
"""

import numpy as np
from numpy import inf, errstate

name = "porod"
title = "Porod function"
description = """\
          I(q) = scale/q^4 + background
"""

category = "shape-independent"

parameters = []

def Iq(q):
    """
    @param q: Input q-value
    """
    with errstate(divide='ignore'):
        return q**-4

Iq.vectorized = True  # Iq accepts an array of q values

def random():
    """Return a random parameter set for the model."""
    sld, solvent = np.random.uniform(-0.5, 12, size=2)
    radius = 10**np.random.uniform(1, 4.7)
    Vf = 10**np.random.uniform(-3, -1)
    scale = 1e-4 * Vf * 2*np.pi*(sld-solvent)**2/(3*radius)
    pars = dict(
        scale=scale,
    )
    return pars

tests = [
    [{'scale': 0.00001, 'background':0.01}, 0.04, 3.916250],
    [{}, 0.0, inf],
]



#!/usr/bin/env python
"""
Evaluate the scattering for a shape using spherical integration over the entire
surface in theta-phi polar coordinates, and compare it to the 1D scattering
pattern for that shape.

Parameters are the same as sascomp.  Unlike the -sphere option in sascomp,
the evaluation is restricted to a single radial line for performance reasons,
with angle set by -angle=alpha in the qx-qy plane.
"""

from __future__ import print_function, division

import sys, os
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

import numpy as np

from sasmodels import compare, data

def main(angle=0, steps=76):
    # Parse the options using the parser in sasmodels.compare.  -angle
    # is an additional parameter that is parsed separately., with -sphere=n and -angle=a as additional arguments
    # Pull out angle ans
    argv = []
    for arg in sys.argv[1:]:
        if arg.startswith('-sphere='):
            steps = int(arg[8:])
        elif arg.startswith('-angle='):
            angle = float(arg[7:])
        elif arg != "-2d":
            argv.append(arg)
    opts = compare.parse_opts(argv)

    # Create a 2D radial slice
    qr = opts['data'].x
    qx, qy = qr*np.cos(np.radians(angle)), qr*np.sin(np.radians(angle))
    radial_data = data.Data2D(x=qx, y=qy)
    radial_data.radial = True  # let plotter know it is actual 1D

    # Create an engine to evaluate it
    comp = compare.make_engine(opts['info'][0], radial_data,
                               opts['engine'][0], opts['cutoff'][0])
    opts['engines'] = [opts['engines'][0], comp]

    # Set the integration parameters to the half sphere
    compare.set_spherical_integration_parameters(opts, steps)

    # Set the random seed
    if opts['seed'] > -1:
        print("Randomize using -random=%i"%opts['seed'])
        np.random.seed(opts['seed'])

    # Run the comparison
    compare.compare(opts, maxdim=np.inf)

if __name__ == "__main__":
    main(angle=0)



r"""
Definition
----------

This model fits the Guinier function

.. math::

    I(q) = \text{scale} \cdot \exp{\left[ \frac{-Q^2 R_g^2 }{3} \right]}
            + \text{background}

to the data directly without any need for linearization (*cf*. the usual
plot of $\ln I(q)$ vs $q^2$\ ). Note that you may have to restrict the data
range to include small q only, where the Guinier approximation actually
applies. See also the guinier_porod model.

For 2D data the scattering intensity is calculated in the same way as 1D,
where the $q$ vector is defined as

.. math:: q = \sqrt{q_x^2 + q_y^2}

In scattering, the radius of gyration $R_g$ quantifies the object's
distribution of SLD (not mass density, as in mechanics) from the object's
SLD centre of mass. It is defined by

.. math:: R_g^2 = \frac{\sum_i\rho_i\left(r_i-r_0\right)^2}{\sum_i\rho_i}

where $r_0$ denotes the object's SLD centre of mass and $\rho_i$ is the SLD at
a point $i$.

Notice that $R_g^2$ may be negative (since SLD can be negative), which happens
when a form factor $P(Q)$ is increasing with $Q$ rather than decreasing. This
can occur for core/shell particles, hollow particles, or for composite
particles with domains of different SLDs in a solvent with an SLD close to the
average match point. (Alternatively, this might be regarded as there being an
internal inter-domain "structure factor" within a single particle which gives
rise to a peak in the scattering).

To specify a negative value of $R_g^2$ in SasView, simply give $R_g$ a negative
value ($R_g^2$ will be evaluated as $R_g |R_g|$). Note that the physical radius
of gyration, of the exterior of the particle, will still be large and positive.
It is only the apparent size from the small $Q$ data that will give a small or
negative value of $R_g^2$.

References
----------

#. A Guinier and G Fournet, *Small-Angle Scattering of X-Rays*,
   John Wiley & Sons, New York (1955)

Authorship and Verification
----------------------------

* **Author:**
* **Last Modified by:**
* **Last Reviewed by:**
"""

import numpy as np
from numpy import inf

name = "guinier"
title = ""
description = """
 I(q) = scale.exp ( - rg^2 q^2 / 3.0 )

    List of default parameters:
    scale = scale
    rg = Radius of gyration
"""
category = "shape-independent"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["rg", "Ang", 60.0, [-inf, inf], "", "Radius of Gyration"]]

source = ["guinier.c"] 


def random():
    """Return a random parameter set for the model."""
    scale = 10**np.random.uniform(1, 4)
    # Note: compare.py has Rg cutoff of 1e-30 at q=1 for Guinier, so use that
    # log_10 Ae^(-(q Rg)^2/3) = log_10(A) - (q Rg)^2/ (3 ln 10) > -30
    #   => log_10(A) > Rg^2/(3 ln 10) - 30
    q_max = 1.0
    rg_max = np.sqrt(90*np.log(10) + 3*np.log(scale))/q_max
    rg = 10**np.random.uniform(0, np.log10(rg_max))
    pars = dict(
        #background=0,
        scale=scale,
        rg=rg,
    )
    return pars

# parameters for unit tests
tests = [[{'rg' : 31.5}, 0.005, 0.992756]]



"""
Execution kernel interface
==========================

:class:`KernelModel` defines the interface to all kernel models.
In particular, each model should provide a :meth:`KernelModel.make_kernel`
call which returns an executable kernel, :class:`Kernel`, that operates
on the given set of *q_vector* inputs.  On completion of the computation,
the kernel should be released, which also releases the inputs.
"""

from __future__ import division, print_function

# pylint: disable=unused-import
try:
    from typing import List, Any
except ImportError:
    pass
else:
    import numpy as np
    from .details import CallDetails
    from .modelinfo import ModelInfo
# pylint: enable=unused-import


class KernelModel(object):
    """
    Model definition for the compute engine.
    """
    info = None  # type: ModelInfo
    dtype = None # type: np.dtype
    def make_kernel(self, q_vectors):
        # type: (List[np.ndarray]) -> "Kernel"
        """
        Instantiate a kernel for evaluating the model at *q_vectors*.
        """
        raise NotImplementedError("need to implement make_kernel")

    def release(self):
        # type: () -> None
        """
        Free resources associated with the kernel.
        """
        #print("null release model")
        pass


class Kernel(object):
    """
    Instantiated model for the compute engine, applied to a particular *q*.

    Subclasses should define *__init__()* to set up the kernel inputs, and
    *_call_kernel()* to evaluate the kernel::

        def __init__(self, ...):
            ...
            self.q_input = <q-value class with nq attribute>
            self.info = <ModelInfo object>
            self.dim = <'1d' or '2d'>
            self.dtype = <kernel.dtype>
            size = 2*self.q_input.nq+4 if self.info.have_Fq else self.q_input.nq+4
            size = size + <extra padding if needed for kernel>
            self.result = np.empty(size, dtype=self.dtype)

        def _call_kernel(self, call_details, values, cutoff, magnetic,
                        radius_effective_mode):
            # type: (CallDetails, np.ndarray, np.ndarray, float, bool, int) -> None
            ... # call <kernel>
            nq = self.q_input.nq
            if self.info.have_Fq:  # models that compute both F and F^2
                end = 2*nq if have_Fq else nq
                self.result[0:end:2] = F**2
                self.result[1:end:2] = F
            else:
                end = nq
                self.result[0:end] = Fsq
            self.result[end + 0] = total_weight
            self.result[end + 1] = form_volume
            self.result[end + 2] = shell_volume
            self.result[end + 3] = radius_effective
    """
    #: Kernel dimension, either "1d" or "2d".
    dim = None  # type: str
    #: Model info.
    info = None  # type: ModelInfo
    #: Numerical precision for the computation.
    dtype = None  # type: np.dtype
    #: Q values at which the kernel is to be evaluated.
    q_input = None  # type: Any
    #: Place to hold result of *_call_kernel()* for subclass.
    result = None # type: np.ndarray

    def Iq(self, call_details, values, cutoff, magnetic):
        # type: (CallDetails, np.ndarray, np.ndarray, float, bool) -> np.ndarray
        r"""
        Returns I(q) from the polydisperse average scattering.

        .. math::

            I(q) = \text{scale} \cdot P(q) + \text{background}

        With the correct choice of model and contrast, setting *scale* to
        the volume fraction $V_f$ of particles should match the measured
        absolute scattering.  Some models (e.g., vesicle) have volume fraction
        built into the model, and do not need an additional scale.
        """
        _, F2, _, shell_volume, _ = self.Fq(call_details, values, cutoff,
                                            magnetic, radius_effective_mode=0)
        combined_scale = values[0]/shell_volume
        background = values[1]
        return combined_scale*F2 + background
    __call__ = Iq

    def Fq(self, call_details, values, cutoff, magnetic,
           radius_effective_mode=0):
        # type: (CallDetails, np.ndarray, np.ndarray, float, bool, int) -> np.ndarray
        r"""
        Returns <F(q)>, <F(q)^2>, effective radius, shell volume and
        form:shell volume ratio.  The <F(q)> term may be None if the
        form factor does not support direct computation of $F(q)$

        $P(q) = <F^2(q)>/<V>$ is used for structure factor calculations,

        .. math::

            I(q) = \text{scale} \cdot P(q) \cdot S(q) + \text{background}

        For the beta approximation, this becomes

        .. math::

            I(q) = \text{scale} P (1 + <F>^2/<F^2> (S - 1)) + \text{background}
                 = \text{scale}/<V> (<F^2> + <F>^2 (S - 1)) + \text{background}

        $<F(q)>$ and $<F^2(q)>$ are averaged by polydispersity in shape
        and orientation, with each configuration $x_k$ having form factor
        $F(q, x_k)$, weight $w_k$ and volume $V_k$.  The result is:

        .. math::

            P(q)=\frac{\sum w_k F^2(q, x_k) / \sum w_k}{\sum w_k V_k / \sum w_k}

        The form factor itself is scaled by volume and contrast to compute the
        total scattering.  This is then squared, and the volume weighted
        F^2 is then normalized by volume F.  For a given density, the number
        of scattering centers is assumed to scale linearly with volume.  Later
        scaling the resulting $P(q)$ by the volume fraction of particles
        gives the total scattering on an absolute scale. Most models
        incorporate the volume fraction into the overall scale parameter.  An
        exception is vesicle, which includes the volume fraction parameter in
        the model itself, scaling $F$ by $\surd V_f$ so that the math for
        the beta approximation works out.

        By scaling $P(q)$ by total weight $\sum w_k$, there is no need to make
        sure that the polydisperisity distributions normalize to one.  In
        particular, any distibution values $x_k$ outside the valid domain
        of $F$ will not be included, and the distribution will be implicitly
        truncated.  This is controlled by the parameter limits defined in the
        model (which truncate the distribution before calling the kernel) as
        well as any region excluded using the *INVALID* macro defined within
        the model itself.

        The volume used in the polydispersity calculation is the form volume
        for solid objects or the shell volume for hollow objects.  Shell
        volume should be used within $F$ so that the normalizing scale
        represents the volume fraction of the shell rather than the entire
        form.  This corresponds to the volume fraction of shell-forming
        material added to the solvent.

        The calculation of $S$ requires the effective radius and the
        volume fraction of the particles.  The model can have several
        different ways to compute effective radius, with the
        *radius_effective_mode* parameter used to select amongst them.  The
        volume fraction of particles should be determined from the total
        volume fraction of the form, not just the shell volume fraction.
        This makes a difference for hollow shapes, which need to scale
        the volume fraction by the returned volume ratio when computing $S$.
        For solid objects, the shell volume is set to the form volume so
        this scale factor evaluates to one and so can be used for both
        hollow and solid shapes.
        """
        self._call_kernel(call_details, values, cutoff, magnetic,
                          radius_effective_mode)
        #print("returned",self.q_input.q, self.result)
        nout = 2 if self.info.have_Fq and self.dim == '1d' else 1
        total_weight = self.result[nout*self.q_input.nq + 0]
        # Note: total_weight = sum(weight > cutoff), with cutoff >= 0, so it
        # is okay to test directly against zero.  If weight is zero then I(q),
        # etc. must also be zero.
        if total_weight == 0.:
            total_weight = 1.
        # Note: shell_volume == form_volume for solid objects
        form_volume = self.result[nout*self.q_input.nq + 1]/total_weight
        shell_volume = self.result[nout*self.q_input.nq + 2]/total_weight
        radius_effective = self.result[nout*self.q_input.nq + 3]/total_weight
        if shell_volume == 0.:
            shell_volume = 1.
        F1 = (self.result[1:nout*self.q_input.nq:nout]/total_weight
              if nout == 2 else None)
        F2 = self.result[0:nout*self.q_input.nq:nout]/total_weight
        return F1, F2, radius_effective, shell_volume, form_volume/shell_volume

    def release(self):
        # type: () -> None
        """
        Free resources associated with the kernel instance.
        """
        #print("null release kernel")
        pass

    def _call_kernel(self, call_details, values, cutoff, magnetic,
                     radius_effective_mode):
        # type: (CallDetails, np.ndarray, np.ndarray, float, bool, int) -> None
        """
        Call the kernel.  Subclasses defining kernels for particular execution
        engines need to provide an implementation for this.
        """
        raise NotImplementedError()



r"""
Definition
----------
This model is a micromagnetic approach to analyse the SANS that arises from
nanoscale variations in the magnitude and orientation of the magnetization in
bulk ferromagnets in the approach to magnetic saturation (single domain state).
Typical materials are cold-worked elemental magnets, hard and soft magnetic 
nanocomposites, amorphous alloys and precipitates in magnetic steel [#Michels2014]_.
The magnetic SANS depends on the magnetic interactions, the magnetic microstructure
(defect/particle size, magnetocrystalline anisotropy, saturation magnetisation)
and on the applied magnetic field. As shown in [#Michels2016]_ near magnetic 
saturation the scattering cross-section can be evaluated by means of micromagnetic theory

.. math::
    I(\mathbf{Q}) = I_{nuc} + I_{mag}(\mathbf{Q},H),

with the field-independent nuclear and magnetic SANS cross section (due
to nanoscale spatial variations of the magnetisation).

.. math::
    I_{mag}(\mathbf{Q},H)= S_K(Q) R_K(\mathbf{Q}, H_i) + S_M(Q) R_M(\mathbf{Q}, H_i),

with $H_i$ the internal field, i.e. the external magnetic field corrected for
demagnetizing effects and the influence of the magnetodipolar field and of the
magnetic anisotropy [#Bick2013]_. This magnetic field dependence of the scattering 
reflects the increasing magnetisation misalignment with decreasing
externally applied magnetic field with a contribution $S_K \times R_K$ due to
perturbations around magnetic anisotropy fields and a term $S_M \times R_M$
related to magnetostatic fields. The magnetic moments decorate perturbations in the
microstructure (precipitates, grain boundaries etc).
The anisotropy-field function $S_K$ depends on the Fourier transform of the magnetic
anisotropy distribution (strength and orientation) in the material, and the
scattering function of the longitudinal magnetisation $S_M$ reflects the
variations of the saturation magnetisation, e.g. jumps at the particle-matrix
interface. $R_K$ and $R_M$ denote the micromagnetic response functions that
describe the magnetisation distribution around a perturbation in magnetic
anisotropy and flucutations in the saturation magnetisation value.

.. figure:: img/micromagnetic_FF.png

    Magnetisation distribution around (left) a particle with magnetic easy axis
    in the vertical direction and (right) a precipitation with a magnetisation
    that is higher than the matrix phase.  

The micromagnetic response functions depend on magnetic material parameters $M_S$:
average saturation magnetisation of the material, $H_i$: the internal magnetic
field, $A$ the average exchange-stiffness constant. In the vicinity of lattice 
imperfection in ferromagnetic materials, antisymmetric Dzyaloshinskii–Moriya
interaction (DMI) can occur due to the local structural inversion symmetry
breaking [#Arrott1963]_. DMI with strength $D$ can give rise to nonuniform spin
textures resulting in a polarization-dependent asymmetric scattering term for 
polycrystalline ferromagnetic with a centrosymmetric crystal structure [#Michels2016]_.
We assume (for simplicity) an isotropic microstructure (for $S_M$) and random 
orientation of magnetic easy axes (additionally for $S_K$) such that the 
contributions of the magnetic microstructure only depend on the magnitude of $q$.
Considerations for a microstructure with a prefered orientation (texture) can be
found in [#Weissmueller2001]_. In the code the averaging procedure over the random
anisotropy is explicitely performed. A specific orientation distribution can be
implemented by rewriting the model. 

The magnetic field is oriented with an inclination of $\alpha$ to the neutron beam
and rotated by $\beta$. The model for the nuclear scattering amplitude, saturation
magnetisation is based on spherical particles with a core shell structure. For
simplicity, only the core has an effective anisotropy, that is varying randomly
in direction from particle to particle. The effect of different, more complex 
spatial profiles of the anisotropy can be seen in [#Michels2010]_.
The magnetic scattering length density (SLD) is defined as 
$\rho_{\mathrm{mag}}=b_H M_S$, where $b_H= 2.91*10^{8}A^{-1}m^{-1}$ and $M_S$
is the saturation magnetisation (in $A/m$).

The fraction of "upward" neutrons before ('up_frac_i') and after the sample 
('up_frac_f') must range between 0 to 1, with 0.5 denoting an unpolarised beam.
Note that a fit may result in a negative magnetic SLD, and hence magnetisation, 
when the polarisation state is inverted, i.e. if you have analysed for a $I_{00}$
state wheras your data are $I_{11}$. The model allows to construct the 4 
spin-resolved cross sections (non-spin-flip $I_{00}$, $I_{11}$ and spin-flip, here
$I_{01}=I_{10}$), half-polarised SANS (SANSpol, incoming polarised beam $I_0$ and
$I_1$, no analysis after sample 'up_frac_f'$=0.5$), and unpolarised beam 
('up_frac_i'$=$'up_frac_f'$=0.5$). Differences and other combinations between
polarised scattering cross section, e.g. to obtain the nuclear-magnetic 
interference scattering, or subtraction of the residual scattering of the high
field reference state can be constructed with a custom model (Fitting>
Add/Multiply Model) and using approbriate scales. For dense systems, special
care has to be taken as the nculear structure factor (arrangement of particles)
does not need to be identical with the magnetic microstructure e.g. local 
textures and correlations between easy axes (see [#Honecker2020]_ for further
details). The use of structure model is therefore strongly discouraged. Better
$I_{nuc}$, $S_K$ and $S_M$ are fit independent from each other in a model-free way. 



References
----------

.. [#Arrott1963] A. Arrott, J. Appl. Phys. 34, 1108 (1963).
.. [#Weissmueller2001] J. Weissmueller et al., *Phys. Rev. B* 63, 214414 (2001).
.. [#Bick2013] J.-P. Bick et al., *Appl. Phys. Lett.* 102, 022415 (2013).
.. [#Michels2010] A. Michels et al., *Phys. Rev. B* 82, 024433 (2010).
.. [#Michels2014] A. Michels, *J. Phys.: Condens. Matter* 26, 383201 (2014).
.. [#Michels2016] A. Michels et al., *Phys. Rev. B* 94, 054424 (2016).
.. [#Honecker2020] D. Honecker, L. Fernandez Barguin, and P. Bender, *Phys. Rev. B* 101, 134401 (2020).



Authorship and Verification
----------------------------

* **Author:** Dirk Honecker **Date:** January 14, 2021
* **Last Modified by:** Dirk Honecker **Date:** September 23, 2024
* **Last Reviewed by:**

"""

import numpy as np
from numpy import pi, inf

name = "micromagnetic_FF_3D"
title = "Field-dependent magnetic microstructure around imperfections in bulk ferromagnets"
description = """
    I(q) = A (F_N^2(q)+ C F_N F_M + D F_M^2) +B(H) I_mag(q,H) 
            A: weighting function =1 for unpolarised beam and non-neutron-spin-flip scattering, zero for spin-flip scattering. The terms in the bracket are the residual scattering at perfectly saturating magnetic field.
            B(H): weighting function for purely magnetic scattering I_mag(q,H)  due to misaligned magnetic moments, different for the various possible spin-resolved scattering cross sections
            C: weighting function for nuclear-magnetic interference scattering
            F_N: nuclear form factor
            F_M: magnetic form factor
The underlying defect can have a core-shell structure.
"""
category = "shape:sphere"

# pylint: disable=bad-whitespace, line-too-long
#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["radius", "Ang", 50., [0, inf], "volume", "Structural radius of the core"],
              ["thickness", "Ang", 40., [0, inf], "volume", "Structural thickness of shell"],            
              ["nuc_sld_core", "1e-6/Ang^2", 1.0, [-inf, inf], "", "Core scattering length density"],
              ["nuc_sld_shell", "1e-6/Ang^2", 1.7, [-inf, inf], "", "Scattering length density of shell"],
              ["nuc_sld_solvent", "1e-6/Ang^2", 6.4, [-inf, inf], "", "Solvent scattering length density"],
              ["mag_sld_core",    "1e-6/Ang^2", 1.0,  [-inf, inf], "",    "Magnetic scattering length density of core"],
              ["mag_sld_shell", "1e-6/Ang^2", 1.7, [-inf, inf], "", "Magnetic scattering length density of shell"],
              ["mag_sld_solvent", "1e-6/Ang^2", 3.0,  [-inf, inf], "",  "Magnetic scattering length density of solvent"],
              ["hk_sld_core", "1e-6/Ang^2", 1.0,  [0, inf], "",    "Anisotropy field of defect"],
              ["Hi", "T", 2.0,  [0, inf], "",    "Effective field inside the material"],
              ["Ms", "T", 1.0,  [0, inf], "",    "Volume averaged saturation magnetisation"], 
              ["A", "pJ/m", 10.0,  [0, inf], "",    "Average exchange stiffness constant"],  
              ["D", "mJ/m^2", 0.0,  [0, inf], "",    "Average DMI constant"],                          
              ["up_i",          "None",  0.5, [0, 1], "", "Polarisation incoming beam"],
              ["up_f",          "None",  0.5, [0, 1], "", "Polarisation outgoing beam"],
              ["alpha",          "None",  90, [0, 180], "", "Inclination of field to neutron beam"],
              ["beta",          "None",  0, [0, 360], "", "Rotation of field around neutron beam"],             
             ]
# pylint: enable=bad-whitespace, line-too-long




source = ["lib/sas_3j1x_x.c", "lib/core_shell.c", "lib/gauss76.c", "lib/magnetic_functions.c", "micromagnetic_FF_3D.c"]
structure_factor = False
have_Fq = False
single=False
opencl = False


def random():
    """Return a random parameter set for the model."""
    outer_radius = 10**np.random.uniform(1.3, 4.3)
    # Use a distribution with a preference for thin shell or thin core
    # Avoid core,shell radii < 1
    radius = np.random.beta(0.5, 0.5)*(outer_radius-2) + 1
    thickness = outer_radius - radius
    pars = dict(
        radius=radius,
        thickness=thickness,
        
    )
    return pars



tests = [
     [{},1.002266990452620e-03, 7.461046163627724e+03],
     [{},(0.0688124,  -0.0261013),  22.024],
]
